/*
*  Copyright (C) 1998-2022 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/
(function() {
  var $ = go.GraphObject.make;

  var staticTests = new TestCollection('Basic');
  TestRoot.add(staticTests);

  staticTests.add(new Test('Existence of static props', null,
    function (test) {
      var diagram = test.diagram;
      var d = diagram;
      diagram.reset();
    }, // END SETUP
    null,
    function (test) {

      test.assert(go.Brush.Solid !== undefined);
      test.assert(go.GraphObject.None !== undefined);
      test.assert(go.GraphObject.Default !== undefined);
      test.assert(go.GraphObject.Vertical !== undefined);
      test.assert(go.GraphObject.Horizontal !== undefined);
      test.assert(go.GraphObject.Fill !== undefined);
      test.assert(go.GraphObject.Uniform !== undefined);
      test.assert(go.GraphObject.make !== undefined);


      test.assert(go.Margin.parse !== undefined);

      test.assert(go.Picture.clearCache !== undefined);
      test.assert(go.Rect.parse !== undefined)
      test.assert(go.Rect.stringify !== undefined)
      test.assert(go.Rect.contains !== undefined)
      test.assert(go.Point.distanceLineSegmentSquared !== undefined);
      test.assert(go.Point.distanceSquared !== undefined);
      test.assert(go.Point.direction !== undefined);
      test.assert(go.Rect.intersects !== undefined)

      /* we dont expose these
      test.assert(go.Margin.ZeroMargin !== undefined);
      test.assert(go.Point.Origin !== undefined);
      test.assert(go.Point.NoPoint !== undefined);
      test.assert(go.Rect.ZeroRect !== undefined)
      test.assert(go.Rect.NoRect !== undefined)
      test.assert(go.Size.ZeroSize !== undefined)
      test.assert(go.Size.OneSize !== undefined)
      test.assert(go.Size.SixSize !== undefined)
      test.assert(go.Size.EightSize !== undefined)
      test.assert(go.Size.TenSize !== undefined)
      test.assert(go.Size.InfiniteSize !== undefined)
      test.assert(go.Size.NoSize !== undefined)
      */

      test.assert(go.Size.parse !== undefined)
      test.assert(go.Size.stringify !== undefined)
      test.assert(go.Spot.NotLeftSide !== undefined)
      test.assert(go.Spot.NotRightSide !== undefined)
      test.assert(go.Spot.NotBottomSide !== undefined)
      test.assert(go.Spot.AllSides !== undefined)
      test.assert(go.Spot.parse !== undefined)
      test.assert(go.Spot.stringify !== undefined)
      test.assert(go.TextBlock.getEllipsis !== undefined);
      test.assert(go.TextBlock.setEllipsis !== undefined);
      test.assert(go.TextBlock.isValidFont !== undefined);
      test.assert(go.Model.fromJSON !== undefined);
      test.assert(go.Model.fromJson === go.Model.fromJSON);
      test.assert(go.Model.fromJson === go.Model.fromJSON);

      test.assert(typeof go.Diagram.prototype.makeSVG === "function")
      test.assert(typeof go.Diagram.prototype.makeSvg === "function")
      test.assert(typeof go.Model.prototype.applyIncrementalJson === "function")
      test.assert(typeof go.Model.prototype.applyIncrementalJSON === "function")


    }


  )); // end test

  staticTests.add(new Test('Proper classNames', null,
    function (test) {
      var diagram = test.diagram;
      diagram.reset();
    }, // END SETUP
    null,
    function (test) {
      function assertName(cls, name) {
        test.assert(cls.className === name, 'Wrong classname for ' + name);
      }

      for (var cls in go) {
        if (go.hasOwnProperty(cls)) {
          switch (cls) {
            case 'Group': assertName(go[cls], 'Group'); break;
            case 'Util': assertName(go[cls], 'Util'); break;
            case 'EnumValue': assertName(go[cls], 'EnumValue'); break;
            case 'List': assertName(go[cls], 'List'); break;
            case 'Set': assertName(go[cls], 'Set'); break;
            case 'Map': assertName(go[cls], 'Map'); break;
            case 'Point': assertName(go[cls], 'Point'); break;
            case 'Size': assertName(go[cls], 'Size'); break;
            case 'Rect': assertName(go[cls], 'Rect'); break;
            case 'Margin': assertName(go[cls], 'Margin'); break;
            case 'Spot': assertName(go[cls], 'Spot'); break;
            case 'Geometry': assertName(go[cls], 'Geometry'); break;
            case 'PathFigure': assertName(go[cls], 'PathFigure'); break;
            case 'PathSegment': assertName(go[cls], 'PathSegment'); break;
            case 'InputEvent': assertName(go[cls], 'InputEvent'); break;
            case 'DiagramEvent': assertName(go[cls], 'DiagramEvent'); break;
            case 'ChangedEvent': assertName(go[cls], 'ChangedEvent'); break;
            case 'Model': assertName(go[cls], 'Model'); break;
            case 'GraphLinksModel': assertName(go[cls], 'GraphLinksModel'); break;
            case 'TreeModel': assertName(go[cls], 'TreeModel'); break;
            case 'GraphModel': assertName(go[cls], 'GraphModel'); break;
            case 'Binding': assertName(go[cls], 'Binding'); break;
            case 'Transaction': assertName(go[cls], 'Transaction'); break;
            case 'UndoManager': assertName(go[cls], 'UndoManager'); break;
            case 'CommandHandler': assertName(go[cls], 'CommandHandler'); break;
            case 'Tool': assertName(go[cls], 'Tool'); break;
            case 'DraggingTool': assertName(go[cls], 'DraggingTool'); break;
            case 'DraggingInfo': assertName(go[cls], 'DraggingInfo'); break;
            case 'DraggingOptions': assertName(go[cls], 'DraggingOptions'); break;
            case 'LinkingBaseTool': assertName(go[cls], 'LinkingBaseTool'); break;
            case 'LinkingTool': assertName(go[cls], 'LinkingTool'); break;
            case 'RelinkingTool': assertName(go[cls], 'RelinkingTool'); break;
            case 'LinkReshapingTool': assertName(go[cls], 'LinkReshapingTool'); break;
            case 'ResizingTool': assertName(go[cls], 'ResizingTool'); break;
            case 'RotatingTool': assertName(go[cls], 'RotatingTool'); break;
            case 'ClickSelectingTool': assertName(go[cls], 'ClickSelectingTool'); break;
            case 'ActionTool': assertName(go[cls], 'ActionTool'); break;
            case 'ClickCreatingTool': assertName(go[cls], 'ClickCreatingTool'); break;
            case 'HTMLInfo': assertName(go[cls], 'HTMLInfo'); break;
            case 'ContextMenuTool': assertName(go[cls], 'ContextMenuTool'); break;
            case 'DragSelectingTool': assertName(go[cls], 'DragSelectingTool'); break;
            case 'PanningTool': assertName(go[cls], 'PanningTool'); break;
            case 'TextEditingTool': assertName(go[cls], 'TextEditingTool'); break;
            case 'ToolManager': assertName(go[cls], 'ToolManager'); break;
            case 'Animation': assertName(go[cls], 'Animation'); break;
            case 'AnimationManager': assertName(go[cls], 'AnimationManager'); break;
            case 'AnimationTrigger': assertName(go[cls], 'AnimationTrigger'); break;
            case 'Layer': assertName(go[cls], 'Layer'); break;
            case 'Diagram': assertName(go[cls], 'Diagram'); break;
            case 'Palette': assertName(go[cls], 'Palette'); break;
            case 'Overview': assertName(go[cls], 'Overview'); break;
            case 'Brush': assertName(go[cls], 'Brush'); break;
            case 'GraphObject': assertName(go[cls], 'GraphObject'); break;
            case 'Panel': assertName(go[cls], 'Panel'); break;
            case 'RowColumnDefinition': assertName(go[cls], 'RowColumnDefinition'); break;
            case 'Shape': assertName(go[cls], 'Shape'); break;
            case 'TextBlock': assertName(go[cls], 'TextBlock'); break;
            case 'TextBlockMetrics': assertName(go[cls], 'TextBlockMetrics'); break;
            case 'Picture': assertName(go[cls], 'Picture'); break;
            case 'Part': assertName(go[cls], 'Part'); break;
            case 'Adornment': assertName(go[cls], 'Adornment'); break;
            case 'Node': assertName(go[cls], 'Node'); break;
            case 'Link': assertName(go[cls], 'Link'); break;
            case 'Placeholder': assertName(go[cls], 'Placeholder'); break;
            case 'Layout': assertName(go[cls], 'Layout'); break;
            case 'LayoutNetwork': assertName(go[cls], 'LayoutNetwork'); break;
            case 'LayoutVertex': assertName(go[cls], 'LayoutVertex'); break;
            case 'LayoutEdge': assertName(go[cls], 'LayoutEdge'); break;
            case 'GridLayout': assertName(go[cls], 'GridLayout'); break;
            case 'PanelLayout': assertName(go[cls], 'PanelLayout'); break;
            case 'CircularLayout': assertName(go[cls], 'CircularLayout'); break;
            case 'CircularNetwork': assertName(go[cls], 'CircularNetwork'); break;
            case 'CircularVertex': assertName(go[cls], 'CircularVertex'); break;
            case 'CircularEdge': assertName(go[cls], 'CircularEdge'); break;
            case 'ForceDirectedLayout': assertName(go[cls], 'ForceDirectedLayout'); break;
            case 'ForceDirectedNetwork': assertName(go[cls], 'ForceDirectedNetwork'); break;
            case 'ForceDirectedVertex': assertName(go[cls], 'ForceDirectedVertex'); break;
            case 'ForceDirectedEdge': assertName(go[cls], 'ForceDirectedEdge'); break;
            case 'LayeredDigraphLayout': assertName(go[cls], 'LayeredDigraphLayout'); break;
            case 'LayeredDigraphNetwork': assertName(go[cls], 'LayeredDigraphNetwork'); break;
            case 'LayeredDigraphVertex': assertName(go[cls], 'LayeredDigraphVertex'); break;
            case 'LayeredDigraphEdge': assertName(go[cls], 'LayeredDigraphEdge'); break;
            case 'TreeLayout': assertName(go[cls], 'TreeLayout'); break;
            case 'TreeNetwork': assertName(go[cls], 'TreeNetwork'); break;
            case 'TreeVertex': assertName(go[cls], 'TreeVertex'); break;
            case 'TreeEdge': assertName(go[cls], 'TreeEdge'); break;
            case 'Debug':
            case 'licenseKey':
            case 'version': break;
            default: test.assert(false, 'unknown class: ' + cls); break;
          }
        }
      }
    }
  )); // end test

})(); // End test system


