/*
*  Copyright (C) 1998-2022 by Northwoods Software Corporation. All Rights Reserved.
*
*  Restricted Rights: Use, duplication, or disclosure by the U.S.
*  Government is subject to restrictions as set forth in subparagraph
*  (c) (1) (ii) of DFARS 252.227-7013, or in FAR 52.227-19, or in FAR
*  52.227-14 Alt. III, as applicable.
*
*  This software is proprietary to and embodies the confidential
*  technology of Northwoods Software Corporation. Possession, use, or
*  copying of this software and media is authorized only pursuant to a
*  valid written license from Northwoods or an authorized sublicensor.
*/

import * as go from "../release/go";

// Type definitions for gotest
export const TestRoot: TestCollection;

export class Test {
  constructor(name: string, diagram: go.Diagram | string | null,
    setup?: ((test: Test) => void) | null,
    execute?: ((test: Test) => void) | null,
    check?: ((test: Test) => void) | null);
  public diagram: go.Diagram;
  public getPath(): string;
  public run(): void;
  public log(msg: string): void;
  public error(msg: string): void;
  public assert(pred: boolean, msg?: string): void;
  assertAllNodeLocations(ptArray: Array<go.Point>): void;
  assertAllLinkPoints(testArrayPoints: Array<Array<go.Point>>): void;
  assertLinkPoints(testPoints: Array<go.Point>): void;
  assertAllLinkGeometries(testArrayGeometries: Array<string>): void;
  assertLinkGeometry(link: go.Link, testGeom: go.Geometry | string): void;
  isApprox(x: number, y: number): boolean;
  isApproxPoint(x: go.Point, y: go.Point): boolean;
  isApproxSize(x: go.Size, y: go.Size): boolean;
  isApproxRect(x: go.Rect, y: go.Rect): boolean;
  mouseDown(pt: go.Point, eventprops?: any): void;
  mouseMove(pt: go.Point, eventprops?: any): void;
  mouseUp(pt: go.Point, eventprops?: any): void;
  mouseWheel(delta: number, eventprops?: any): void;
  keyDown(keyorcode: string | number, eventprops?: any): void;
  keyUp(keyorcode: string | number, eventprops?: any): void;
}

export class TestCollection {
  constructor(name: string);
  public add(test: Test | TestCollection): void;
  public RunSlowTests: boolean
}

export function getTree(): go.Diagram;
export function initTree(): void;
export function collapseTree(levels: number): void;
